#define _GNU_SOURCE
#include <stdio.h>
#include <assert.h>

#include <time.h>
#include <unistd.h>

#include "def.h"
#include "conf/conf.h"
#include "utils/utils.h"
#include "libkatrin-db-mysql.h"

static MYSQL mysqllink;

/* Module initialization */
void module_init(void) {
	dbConnect();
}

/* Connect to db
 * return 0 if OK, else error code */
int dbConnect(void) {
	kcfg *cfg = get_katrin_cfg();

	if (cfg == NULL) {
		err("Can't get katrin configuration");
		return 1;
	}

	debug("Connect to DB...");

	mysql_init(&mysqllink);
	my_bool reconnect = 0;
	mysql_options(&mysqllink, MYSQL_OPT_RECONNECT, &reconnect);

	if (!(mysql_real_connect(&mysqllink, cfg->db_server, cfg->db_login, cfg->db_password, cfg->db_database, 0, NULL, 0))) {
		err(" Error connection to DB: %s", mysql_error(&mysqllink));
		free_katrin_cfg(cfg);
		return 2;
	}

	debug(" Good connect to DB");
	free_katrin_cfg(cfg);
	return 0;
}

/* Disconnect from db */
void dbDisconnect(void) {
	mysql_close(&mysqllink);
}

/* Get tariff struct by id
 * @param id: tarrif id
 * return: pointer to tariff struct. NULL if error or tariff not exist */
struct tariff *get_tariff(int id) {
	debug("get_tariff");
	char query[MAX_QUERY_SIZE];
	sprintf(query, "SELECT id, week_days FROM tariffs WHERE id='%d';", id);
	struct tariff *p = NULL;
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;

	if (katrin_mysql_query(query) != 0)
		return NULL;
	res=mysql_store_result(&mysqllink);

	if (mysql_num_rows(res) == 0) {
		warn("Can't get tariff!");
		return NULL;
	}

	row = mysql_fetch_row(res);
	p = Tariff(atoi(row[0]),row[3]);

	mysql_free_result(res);

	return p;
}

/* Get account by id 
 * @param id: account id 
 * return: pointer to account struct */
struct account *get_account(int id) {
	debug("get_account");
	char query[MAX_QUERY_SIZE];
	sprintf(query, "SELECT id, balance, credit, blocked, type, client_id, group_id \
			FROM accounts WHERE id='%d';", id);
	struct account *p = NULL;
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;

	if (katrin_mysql_query(query) != 0)
		return NULL;
	res=mysql_store_result(&mysqllink);

	if (mysql_num_rows(res)==0) {
		debug("Can't get account!");
	} else {
		row = mysql_fetch_row(res);
		p = Account(atoi(row[0]), atof(row[1]), atof(row[2]),\
				atoi(row[3]), atoi(row[4]), atoi(row[5]), atoi(row[6]));
	}

	mysql_free_result(res);

	return p;
}

/* Writeoff cost from account
 * @param cost: cost
 * @param account_id: id of account */
void writeoffcost(double cost, int account_id) {
	debug("writeoffcost");
	char query[MAX_QUERY_SIZE];
	sprintf(query,"UPDATE accounts SET balance=balance-%1.4f WHERE id='%d';", cost, account_id);
	debug("writeoffcost query: %s", query);
	katrin_mysql_query(query);
}

/* Set block account 
 * @param account_id: id of account
 * @param value: blocked bit. 1 if blocked */
void set_account_blocked(int account_id, int value) {
	debug("set_account_blocked");
	char query[MAX_QUERY_SIZE];
	sprintf(query,"UPDATE accounts SET blocked='%d' WHERE id='%d';", value, account_id);
	katrin_mysql_query(query);
}

/* Check login & password
 * @param login: account login
 * @param password: account password
 * return: 0 if login & password valid instead 1 */
int checkloginpassword(const char *login, const char *password) {
	debug("checkloginpassword");
	char query[MAX_QUERY_SIZE];
	sprintf(query,"SELECT id FROM clients WHERE login='%s' AND password='%s';", login, password);
	katrin_mysql_query(query);
	MYSQL_RES *res = mysql_store_result(&mysqllink);
	if (mysql_affected_rows(&mysqllink)>0) {
		mysql_free_result(res);
		return 1;
	}
	mysql_free_result(res);
	return 0;
}

/* list of services associated with groups of given client
 * serviceName!=NULL => returns first service with that name
 * So katrin admin should guarantee uniqueness of services among all client groups
 * @param client_id: client id
 * @param serviceName: name of Service or NULL
 * return NULL if error, else pointer to serviceslist */
struct serviceslist *get_client_clientgroups_services(int client_id, char *serviceName) {
	debug("get_client_clientgroups_services");
	char query[MAX_QUERY_SIZE];
	char andservice[MAX_QUERY_SIZE];
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;

	struct serviceslist *usl = NULL;
	struct serviceslist *t = NULL;
	struct tariff *tar = NULL;
	struct account *ac = NULL;
	if (sprintf(query, "SELECT s.codename, \
				a.id, a.balance, a.credit, a.blocked, a.type, a.group_id, \
				t.id, t.week_days \
			FROM clientgroups_clients as cg INNER JOIN clients as c INNER JOIN \
				services as s INNER JOIN user_services as us INNER JOIN accounts as a INNER JOIN tariffs as t \
			ON cg.client_id = c.id AND us.group_id = cg.clientgroup_id AND \
				s.id = us.service_id AND a.id = us.account_id AND t.id = us.tariff_id \
			WHERE c.id = %d", client_id) < 0 )
	{
		err("sprintf error in get_client_clientgroups_services");
		return NULL;
	};
	if(serviceName) {
		sprintf(andservice," AND s.codename='%s' LIMIT 1", serviceName);
		strcat(query, andservice);
	}

	debug("QUERY: %s", query);
	if (katrin_mysql_query(query) != 0)
		return NULL;
	res = mysql_store_result(&mysqllink);
	int num_rows = mysql_num_rows(res);

	int i;
	for (i = 0; i<num_rows ; i++) {
		row = mysql_fetch_row(res);

		t = (struct serviceslist *)malloc(sizeof(struct serviceslist));

		if (t == NULL)
			err("Can't allocate memory");
		else {
			t->next = usl;
			strcpy(t->service.name, row[0]);
			ac = Account(atoi(row[1]), atof(row[2]), atof(row[3]), \
					atoi(row[4]), atoi(row[5]), NULL , anulltoi(row[6]));
			t->account = *ac;
			free(ac);

			tar = Tariff(atoi(row[7]), row[8]);
			t->tariff = *tar;
			free(tar);
			usl = t;
		}
	}
	mysql_free_result(res);
	debug("exit from get_client_clientgroups_services");
	return usl;
}

/* Get list of services associated with given client_id or group_id
 * type==Group => id is group_id otherwise => id is client_id
 * serviceName!=NULL => return service with that name
 * We assume that there is only one service with this name - MUST be guarantied by database 
 * @param id: client_id or group_id
 * @param accountType: account type
 * @param serviceName: service codename
 * return: NULL on error, else pointer to serviceslist */
struct serviceslist *get_client_services(int id, enum accountType type, char *serviceName) {
	debug("get_client_services");

	char query[MAX_QUERY_SIZE];
	char where[MAX_QUERY_SIZE];
	char andservice[MAX_QUERY_SIZE];
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;

	struct serviceslist *usl = NULL;
	struct serviceslist *t = NULL;
	struct tariff *tar = NULL;
	struct account *ac = NULL;
	if (sprintf(query, "SELECT s.codename, \
				a.id, a.balance, a.credit, a.blocked, a.type, a.client_id, a.group_id, \
				t.id, t.week_days \
			FROM services as s INNER JOIN user_services as us INNER JOIN accounts as a INNER JOIN tariffs as t \
			ON s.id = us.service_id AND a.id = us.account_id AND t.id = us.tariff_id") <0 )
	{
		err("sprintf error in get_client_services");
		return NULL;
	};
	if (type == Group)
		sprintf(where," WHERE us.group_id='%d'", id);
	else
		sprintf(where," WHERE us.client_id='%d'", id);
	if (serviceName) {
		sprintf(andservice," AND s.codename='%s'", serviceName);
		strcat(where, andservice);
	}

	strcat(query, where);
	if (katrin_mysql_query(query) != 0)
		return NULL;
	res = mysql_store_result(&mysqllink);
	int num_rows = mysql_num_rows(res);

	int i;
	for (i = 0; i<num_rows ; i++) {
		row = mysql_fetch_row(res);

		t = (struct serviceslist *)malloc(sizeof(struct serviceslist));

		if (t == NULL)
			err("Can't allocate memory");
		else {
			t->next = usl;
			strcpy(t->service.name, row[0]);
			ac = Account(atoi(row[1]), atof(row[2]), atof(row[3]), \
					atoi(row[4]), atoi(row[5]), anulltoi(row[6]), anulltoi(row[7]));
			t->account = *ac;
			free(ac);

			tar = Tariff(atoi(row[8]), row[9]);
			t->tariff = *tar;
			free(tar);
			usl = t;
		}
	}
	mysql_free_result(res);
	debug("exit from get_client_services");
	return usl;
}

/* Get client struct finded by id or login if id=0 or by param 
 * @param id: id client or 0 
 * @param login: client login 
 * @param typeService: codename service
 * @param uparam: param name
 * @param uvalue: param value 
 * return: NULL if user not exist, else pointer to client struct */
struct client *get_client(int id, char *login, char *typeService, char *uparam, char *uvalue) {
	debug("get_client");
	debug("id=%d login=%s typeService=%s uparam=%s uvalue=%s", id, login, typeService, uparam, uvalue);
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;
	char query[MAX_QUERY_SIZE];
	char where[MAX_QUERY_SIZE];
	struct client *p = NULL;

	/* Search client by id or login */
	int usewhere = 0;
	if (id!=0 || login!=NULL) {
		strcpy(query, "SELECT id,login,password FROM clients ");

		if (id!=0) {
			sprintf(where," WHERE id='%d'", id);
			strcat(query, where);
			usewhere = 1;
		}

		if (login!=NULL) {
			if (usewhere==1)
				strcat(query, " AND");
			else
				strcat(query, " WHERE");
			sprintf(where," login='%s'", login);
			strcat(query, where);
			usewhere = 1;
		}

		debug(query);

		if (katrin_mysql_query(query) == 0) {
			res=mysql_store_result(&mysqllink);
			if (mysql_affected_rows(&mysqllink)==0) {
				debug("exit from get_client with NULL");
				return NULL;
			}
			else {
				row=mysql_fetch_row(res);
				p = (struct client *) calloc(1,sizeof(struct client));
				p->id = atoi(row[0]);
				assert(strlen(row[2]) < sizeof(p->password));
				strcpy(p->password,row[2]);
			}
			mysql_free_result(res);
		}
	}
	/* Search and add client by param or only add params if client was already found */
	int service_added = 0;
	if (typeService && uparam && uvalue) {
		sprintf(query,"SELECT * FROM service_%s_uparams WHERE %s='%s' LIMIT 1", typeService, uparam, uvalue);
		debug(query);
		if (katrin_mysql_query(query) == 0) {
			res = mysql_store_result(&mysqllink);
			MYSQL_ROW_OFFSET saved_pos = mysql_row_tell(res);
			if (mysql_num_rows(res) > 0) {
				row = mysql_fetch_row(res);
				debug("client_id=%s", row[1]);
				if (p == NULL) {
					debug("client is null.. get_client...");
					p = get_client(atoi(row[1]), NULL, NULL, NULL, NULL);
					if (p == NULL)
						debug("client is null too");
					else {
						debug("GOOD get_client");
						service_added = 1;
					}
				}
				else
					debug("client is not null");
				mysql_row_seek(res, saved_pos);
				p->uparams = get_client_params(p->id, typeService);
			}
			else {
				debug("Can't get client with this param");
			}
			mysql_free_result(res);
		}
	}
	else {
		if (typeService)
			p->uparams = get_client_params(p->id, typeService);
		else
			if (p!=NULL)
				p->uparams = NULL;
	}

	/* Add client services */
	if (p!=NULL && service_added == 0) {
		p->services = get_client_services(p->id, Personal, NULL);
	}

	debug("exit from getclient");
	return p;
}

/* Get list of abonents with this blocked state. 
 * Set this blocked state to all accounts
 * @param accountState: state of account
 * return: NULL on error or pointer to abonentslist struct */
struct abonentslist *update_account_state(enum accountState blocked) {
	char query[MAX_QUERY_SIZE];
	char where[MAX_QUERY_SIZE];
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;
	int num_rows = 0;
	struct abonentslist *p;
	// the top of the list
	struct abonentslist *top = NULL;
	int i;

	debug("update_account_state(%d)", blocked);
	if (sprintf(query, "SELECT a.type, us.client_id, us.group_id, s.codename \
			FROM user_services as us INNER JOIN accounts as a INNER JOIN services as s  \
			ON a.id = us.account_id AND s.id = us.service_id") < 0)
	{
		err("sprintf error in update_account_state (query)");
		return NULL;
	};
	if (sprintf(where, " WHERE a.blocked = %d AND a.balance + a.credit %s;", !blocked, blocked?"<= 0":"> 0") < 0)
	{
		err("sprintf error in update_account_state (where)");
		return NULL;
	};
	strncat(query, where, MAX_QUERY_SIZE);

	if (katrin_mysql_query(query) !=0 )
		return NULL;
	res=mysql_store_result(&mysqllink);
	num_rows = mysql_affected_rows(&mysqllink);

	for (i = 0; i < num_rows; i++) {
		row=mysql_fetch_row(res);
		p = (struct abonentslist *) malloc(sizeof(struct abonentslist));
		p->next = top;
		top = p;
		top->type = atoi(row[0]);
		top->abonent = Abonent(anulltoi(row[1]), anulltoi(row[2]),NULL, top->type);
		strncpy(top->service.name, row[3], SERVICE_NAME_LENGTH);
	}
	mysql_free_result(res);
	// Let's update these accounts
	if (sprintf(query, "UPDATE accounts as a SET a.blocked = %d", blocked) < 0)
	{
		err("sprintf error in update_account_state (where, UPDATE)");
		return NULL;
	};
	strncat(query, where, MAX_QUERY_SIZE);

	if (katrin_mysql_query(query) !=0 )
		err("Can't update accounts state");
	return top;
}

/* Get clients in group by client group id
 * @param clientgroup_id: id of client group
 * return: NULL on error, else pointer to clientslist struct */
struct clientslist *get_clientgroup_members(int clientgroup_id) {
	debug("get_clientgroup_members for clientgroup_id %d",clientgroup_id);
	char query[MAX_QUERY_SIZE];
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;
	int num_rows = 0;
	int i;
	struct clientslist *p;
	struct clientslist *top = NULL;
	if (sprintf(query, "SELECT c.id \
			FROM clientgroups_clients as cg INNER JOIN clients as c  \
			ON cg.client_id = c.id \
			WHERE cg.clientgroup_id = %d;", clientgroup_id) < 0)
	{
		err("sprintf error in get_clientgroup_members");
		return NULL;
	};

	if (katrin_mysql_query(query) !=0 )
		return NULL;
	res=mysql_store_result(&mysqllink);
	num_rows = mysql_affected_rows(&mysqllink);

	for (i = 0; i < num_rows; i++) {
		row=mysql_fetch_row(res);
		p = (struct clientslist *) malloc(sizeof(struct clientslist));
		p->next = top;
		top = p;
		top->client = (struct client *) calloc(1,sizeof(struct client));
		top->client->id = atoi(row[0]);
	}
	mysql_free_result(res);
	return top;
}

/* Add service stat record to db
 * @param typeService: service codename
 * @param client_id: client id
 * @param filter_id: filter id
 * @param cost: cost
 * @param paramlist: list of params */
void addstat(char *typeService, int client_id, int filter_id, double cost, struct paramlist *params) {
	debug("addstat");
	time_t now = time(NULL);

	char query[MAX_QUERY_SIZE];

	char field_list[MAX_FIELD_LIST_SIZE];
	char value_list[MAX_VALUE_LIST_SIZE];
	paramlist2strlists(params, field_list, value_list, ",`", "`", ",'", "'");

	sprintf(query, "INSERT INTO service_%s_stats (`client_id`, `filter_id`, `date_time`, `cost`, %s) VALUES ('%d', '%d', FROM_UNIXTIME(%ld), %f, %s);", typeService, field_list, client_id, filter_id, now, cost, value_list);
	katrin_mysql_query(query);
}

/* Function returns filters with tariff_id or all filters if tariff_id==0
 * @param tariff_id: tariff id
 * @param typeService: service codename
 * return: NULL on error, else pointer to filterslist struct */
struct filterslist *tariffid2filters(int tariff_id, char *typeService) {
	debug("tariffid2filters");
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;
	char query[MAX_QUERY_SIZE];
	char where[50] = "\0";

	if (tariff_id>0)
		sprintf(where,"WHERE tariff_id=%d", tariff_id);
	sprintf(query,"SELECT * FROM service_%s_filters %s order by priority;", typeService, where);

	if (katrin_mysql_query(query) !=0 )
		return NULL;

	res = mysql_store_result(&mysqllink);

	int num_rows = mysql_affected_rows(&mysqllink);
	debug("num filters for this tariff: %d", num_rows);
	struct filterslist *p = NULL;
	struct filterslist *ret_p = NULL;
	int i = 0;

	for (i = 0; i<num_rows; i++) {
		if (i==0) {
		   ret_p = (struct filterslist *) malloc(sizeof(struct filterslist));
		   p = ret_p;
		}
		else {
			p->next = (struct filterslist *) malloc(sizeof(struct filterslist));
			p = p->next;
		}

		p->filter.params = NULL;	
		p->filter.params = MySQLres2paramlist(res, 6);

		row = mysql_fetch_row(res);
		p->next = NULL;
		p->filter.id = atoi(row[0]);
		/* row[1] is title. We don't need it */
		p->filter.tariffid = atoi(row[2]);
		/* row[3] is filter_group_id. We don't need it */
		p->filter.priority = atoi(row[4]);
		p->filter.store_stat = atoi(row[5]);
	}

	mysql_free_result(res);
	debug("exit from tariffid2filters");
	return ret_p;
}

/* Function returns clients with tariffid or all clients if tariffid==0
 * @param tariff_id: tariff id
 * return: NULL on error, pointer to clientlist struct */
struct clientslist *tariffid2clients(int tariff_id) {
	debug("tariffid2clients");
	MYSQL_RES *res = NULL;
	MYSQL_ROW row = NULL;
	char query[MAX_QUERY_SIZE];
	int num_rows = 0;
	sprintf(query,"SELECT client_id FROM user_services WHERE tariff_id='%d' and client_id is not NULL;",tariff_id);
	if (katrin_mysql_query(query) !=0)
		return NULL;
	res=mysql_store_result(&mysqllink);
	num_rows = mysql_affected_rows(&mysqllink);

	struct clientslist *p;
	struct clientslist *top = NULL;
	int i;
	for (i = 0; i<num_rows ; i++) {
		row=mysql_fetch_row(res);
		p = (struct clientslist *) malloc(sizeof(struct clientslist));
		p->client = (struct client *) malloc(sizeof(struct client));
		p->client->id = atoi(row[0]);
		p->next = top;
		top = p;
	}
	mysql_free_result(res);
	return top;
}

/* Get client params by client id and service codename 
 * @param client_id: id of client
 * @param typeService: service codename
 * return: param list or NULL on error */
struct paramlist *get_client_params(int client_id, char *typeService) {
	debug("get_client_params");
	MYSQL_RES *res = NULL;
	char query[MAX_QUERY_SIZE];
	sprintf(query,"SELECT * FROM service_%s_uparams WHERE client_id=%d;", typeService, client_id);

	if (katrin_mysql_query(query) != 0)
		return NULL;
	
	res = mysql_store_result(&mysqllink);

	struct paramlist *pl = MySQLres2paramlist(res, 2);
	mysql_free_result(res);
	return pl;
}

/* Get client group params (name)
 * @param clientgroup_id: id of clientgroup
 * @param typeService: codename of service 
 * return: param list */
struct paramlist *get_clientgroup_params(int clientgroup_id, char *typeService) {
	debug("get_clientgroup_params");
	MYSQL_RES *res = NULL;
	char query[MAX_QUERY_SIZE];
	sprintf(query,"SELECT name FROM clientgroups WHERE id=%d;",clientgroup_id);

	if (katrin_mysql_query(query) != 0)
		return NULL;
	
	res = mysql_store_result(&mysqllink);
	struct paramlist *pl = MySQLres2paramlist(res, 0);
	mysql_free_result(res);
	return pl;
}


/* Additional specific function for mysql db */

/* Convert mysql res to param list with start from start_fld index
 * @param res: MYSQL result
 * @ start_fld: start field index
 * return: param list */
static struct paramlist *MySQLres2paramlist(MYSQL_RES *res, int start_fld) {
	debug("MySQLres2paramlist");
	if (mysql_num_rows(res) == 0) {
		debug("Null mysql res for convert to paramlist!");
		return NULL;
	}
	MYSQL_ROW_OFFSET saved_pos = mysql_row_tell(res);
	MYSQL_ROW row = NULL;
	row = mysql_fetch_row(res);
	MYSQL_FIELD *fld;
	struct paramlist *pl = NULL;
	struct paramlist *top = NULL;
	int num_fields = mysql_num_fields(res);
	debug("All num fields: %d", num_fields);

	int j=0;
	for (j = start_fld; j < num_fields; j++) {
		/* XXX: Buffer overflow may occurs if row[j] >= MAX_PARAM_VALUE_LENGTH */
		assert(strlen(row[j]) < MAX_PARAM_VALUE_LENGTH);

		pl = (struct paramlist *) malloc(sizeof(struct paramlist));
		if (pl == NULL) {
			err("Can't allocate memory for paramlist struct");
			free_paramlist(top);
			return NULL;
		}
		pl->next = top;
		top = pl;
		fld = mysql_fetch_field_direct(res, j);
		strncpy(top->param, fld->name, MAX_PARAM_NAME_LENGTH);
		debug("row[%d]=%s; strlen(row[%d])=%d; fld->name=%s", j, row[j], j, strlen(row[j]), fld->name);
		strncpy(top->value, row[j], MAX_PARAM_VALUE_LENGTH);
	}
	mysql_row_seek(res, saved_pos);
	debug("exit from MySQLres2paramlist");
	return top;
}

/* wrapper over mysql_real_query function 
 * @param query: pointer to query string
 * return: 0 if OK, else 1 */
static int katrin_mysql_query(const char *query) {
	debug("katrin_mysql_query");
	int rez = 0;

	while (mysql_ping(&mysqllink) != 0) {
		dbDisconnect();
		debug("attempt connect to DB");
		dbConnect();
		usleep ( 1000000 ) ;
	}

	rez = mysql_real_query(&mysqllink, query, strlen(query));
	if (rez == 0 ) {
		debug("good exit from katrin_mysql_query");
		return rez;
	}
	else {
		err("MySQL query error: %s", mysql_error(&mysqllink));
		debug("QUERY: %s", query);
		rez = 1;
	}

	debug("error exit from katrin_mysql_query");
	return rez;
}


